import torch
import torch.nn as nn
import torch.optim as optim
import torchaudio
import tqdm

from torch.utils.data import random_split
from torch.utils.data import DataLoader

from data import AudioDataset
from generator import SISUGenerator
from discriminator import SISUDiscriminator

# Check for CUDA availability
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDataset(dataset_dir, target_duration=2.0)

dataset_size = len(dataset)
train_size = int(dataset_size * .9)
val_size = int(dataset_size-train_size)

train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

train_data_loader = DataLoader(train_dataset, batch_size=4, shuffle=True)
val_data_loader = DataLoader(val_dataset, batch_size=4, shuffle=True)

# Initialize models and move them to device
generator = SISUGenerator()
discriminator = SISUDiscriminator()

generator = generator.to(device)
discriminator = discriminator.to(device)

# Loss
criterion_g = nn.L1Loss()  # Perceptual Loss (L1 instead of MSE)
criterion_d = nn.MSELoss()  # Can keep MSE for discriminator (optional)

# Optimizers
optimizer_g = optim.Adam(generator.parameters(), lr=0.0001, betas=(0.5, 0.999))  # Reduced learning rate
optimizer_d = optim.Adam(discriminator.parameters(), lr=0.0001, betas=(0.5, 0.999))

# Learning rate scheduler
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_d, mode='min', factor=0.1, patience=5)

# Training loop
num_epochs = 500

for epoch in range(num_epochs):
    low_quality_audio = torch.empty((1))
    high_quality_audio = torch.empty((1))
    ai_enhanced_audio = torch.empty((1))
    total_d_loss = 0
    total_g_loss = 0

    # Training
    for low_quality, high_quality in tqdm.tqdm(train_data_loader, desc=f"Epoch {epoch+1}/{num_epochs}"):
        high_quality = high_quality.to(device)
        low_quality = low_quality.to(device)

        batch_size = high_quality.size(0)
        real_labels = torch.ones(batch_size, 1).to(device)
        fake_labels = torch.zeros(batch_size, 1).to(device)

        ###### Train Discriminator ######
        discriminator.train()
        optimizer_d.zero_grad()

        # 1. Real data
        real_outputs = discriminator(high_quality)
        d_loss_real = criterion_d(real_outputs, real_labels)

        # 2. Fake data
        fake_audio = generator(low_quality)
        fake_outputs = discriminator(fake_audio.detach())  # Detach to stop gradient flow to the generator
        d_loss_fake = criterion_d(fake_outputs, fake_labels)

        d_loss = (d_loss_real + d_loss_fake) / 2.0  # Without gradient penalty
        d_loss.backward()
        optimizer_d.step()
        total_d_loss += d_loss.item()

        generator.train()
        optimizer_g.zero_grad()

        # Generator loss: how well fake data fools the discriminator
        fake_outputs = discriminator(fake_audio)  # No detach here
        g_loss = criterion_g(fake_outputs, real_labels)  # Train generator to produce real-like outputs

        g_loss.backward()
        optimizer_g.step()
        total_g_loss += g_loss.item()

        low_quality_audio = low_quality
        high_quality_audio = high_quality
        ai_enhanced_audio = fake_audio

    if epoch % 10 == 0:
        print(f"Saved epoch {epoch}!")
        torchaudio.save(f"./output/epoch-{epoch}-audio-crap.wav", low_quality_audio[0].cpu(), 44100)
        torchaudio.save(f"./output/epoch-{epoch}-audio-ai.wav", ai_enhanced_audio[0].cpu(), 44100)
        torchaudio.save(f"./output/epoch-{epoch}-audio-orig.wav", high_quality_audio[0].cpu(), 44100)

torch.save(generator.state_dict(), "generator.pt")
torch.save(discriminator.state_dict(), "discriminator.pt")

print("Training complete!")
